
densityPlotsLogPMsandMMs<-function(data,genes=NULL){

  # Density plot curves of logPMs and logMMS.
  # One curve per array, same line-type used for
  # arrays under the same condition

  if(!require(affy))
      stop("The affy package is required and could not be found\n")

  if(is.null(data))
      {
        stop(" parameter 'data'=NULL: You have not provided an affybatych object\n")
      }

  if(is.null(genes))
      {
        cat("parameter 'genes' not specified: basing plots on all genes.\n")
        genes=geneNames(data)
      }
  else
      {
        cat("parameter 'genes' specified: basing plots on this subset of genes.\n")
        genes=geneNames(data)[genes]
      }

  pm=pm(data, genes)
  mm=mm(data, genes)

  samples=(dim(pm(data))[2])
  themax=0
  for(s in 1:samples){
    themax= max(themax,max(log(pm)[,s]))
  }

  par(mfrow=c(1,3))
  plot(density(log(pm)[,1]),xlim=range(0,themax),type="n",main="log(PM)")
  for(s in 1:samples){
    lines(density(log(pm)[,s]),pch=paste(s))
  } 
  plot(density(log(mm)[,1]),xlim=range(0,themax),type="n",main="log(MM)")
  for(s in 1:samples){
    lines(density(log(mm)[,s]),pch=paste(s))
  } 
  diff<-pm[,1]-mm[,1]
  plot(density(log(diff[diff>0])),xlim=range(0,themax),type="n",main="log(PM-MM*)")
  for(s in 1:samples){
    diff=pm[,s]-mm[,s]
    lines(density(log(diff[diff>0])),pch=paste(s))
  } 
}



densityPlotsLogSandHandmu<-function(){

  # density plot curves for log(S+1), log(H+1) and mu.
  # For log(S+1) and log(H+1): one curve per array,
  # for mu: one curve per condition. Different line-types
  # used or different conditions.

  cat("Attempting to load 'bgxFunctionCall.Rdata'.\n")
  cat("If not successfull make sure your working directory is correct.\n")
  load("bgxFunctionCall.Rdata") 
  cat("  'bgxFunctionCall.Rdata' has successfully been loaded.\n")
 
  load("namesOfGenesAnalysed.Rdata")
  genes=namesOfGenesAnalysed

  samples=sum(runParameterList$samplesets)
  probes=(runParameterList$numberProbesAnalysed)
  samplesets=runParameterList$samplesets
  
  help=ceiling(probes/10000.0)
  par(mfrow=c(1,3))

  x<-array(scan("xave"),dim=c(probes,samples))
  plot(density(x[seq(1,probes,by=help),1]),main="log(S+1)",xlim=range(0,10),type="n")
  samplesSoFar=0
  for(c in 1:length(samplesets)){
    for(s in 1:samplesets[c]){
      samplesSoFar=samplesSoFar+1
      lines(density(x[seq(1,probes,by=help),samplesSoFar]),pch=c)
    }
  }

  y<-array(scan("yave"),dim=c(probes,samples))
  plot(density(y[seq(1,probes,by=help),1]),main="log(H+1)",xlim=range(0,10),type="n")
  samplesSoFar=0
  for(c in 1:length(samplesets)){
    for(s in 1:samplesets[c]){
      samplesSoFar=samplesSoFar+1
      lines(density(y[seq(1,probes,by=help),samplesSoFar]),pch=c)
    }
  }

  mu<-array(scan("mugave"),dim=c(length(genes),length(samplesets)))
  plot(density(mu[,1]),main="mu",xlim=range(0,10))
  for(c in 1:length(samplesets))
    lines(density(mu[,c]),pch=c)

  cat("Plotted kernel density curves for log(S+1) and log(H+1)\n")
  cat("(one for each array) and for mu (one for each condition)\n")
}



muMAplots<-function(){

  # MA mu plots: plotting mean(ci)-mean(cj) against 0.5*(mean(ci)+mean(cj))
  # for pairs of conditions (ci,cj), with  cj=1,...,numberConditions
  # and ci=cj+1,...,numberConditions. Fits a loess curve to this
  # MA plot and draws the curve.

  cat("Attempting to load 'bgxFunctionCall.Rdata'.\n")
  cat("If not successfull make sure your working directory is correct.\n")
  load("bgxFunctionCall.Rdata") 
  cat("  'bgxFunctionCall.Rdata' has successfully been loaded.\n")

  load("namesOfGenesAnalysed.Rdata")
  genes=namesOfGenesAnalysed

  samplesets=runParameterList$samplesets

  if(length(samplesets) == 2) help=1
  else  help=ceiling(sqrt(length(samplesets)))

  mu<-array(scan("mugave"),dim=c(length(genes),length(samplesets))) 
  par(mfrow=c(help,help))
  for(c2 in 1:length(samplesets)){
    if( (c2+1) < length(samplesets)+1){
      for(c1 in (c2+1):length(samplesets)){
        plot(0.5*(mu[,c1]+mu[,c2]),mu[,c1]-mu[,c2],main=paste("MAplot for mu: cond ",c1, " against ",c2),xlab="A",ylab=paste("M:",c1,"-",c2),pch=20)
        lines(c(0,15),c(0,0),col="grey")      
        help=loess(M~A,data.frame(M=(mu[,c1]-mu[,c2]),A=0.5*(mu[,c1]+mu[,c2])))  
        predictedVals=predict(help,data.frame(A=0.5*(mu[,c1]+mu[,c2])))
        points(0.5*(mu[,c1]+mu[,c2]),predictedVals,col=3,pch=20)
      }    
    }
  }
  cat("Plotted (A,M) plots with loess lines for \mu for all pairswise comparisons\n")
}


muMAplotsAfterLoess<-function(){

  # plotting mean(ci)-mean(cj)-loessCurve fit to MA plot
  # against 0.5*(mean(ci)+mean(cj)) for pairs of conditions (ci,cj), 
  # with  cj=1,...,numberConditions and ci=cj+1,...,numberConditions. 

  cat("Attempting to load 'bgxFunctionCall.Rdata'.\n")
  cat("If not successfull make sure your working directory is correct.\n")
  load("bgxFunctionCall.Rdata") 
  cat("  'bgxFunctionCall.Rdata' has successfully been loaded.\n")

  load("namesOfGenesAnalysed.Rdata")
  genes=namesOfGenesAnalysed
  samplesets=runParameterList$samplesets

  if(length(samplesets) == 2) help=1
  else  help=ceiling(sqrt(length(samplesets)))

  mu<-array(scan("mugave"),dim=c(length(genes),length(samplesets))) 
  par(mfrow=c(help,help))
  for(c2 in 1:length(samplesets)){
    if( (c2+1) < length(samplesets)+1){
      for(c1 in (c2+1):length(samplesets)){
        help=loess(M~A,data.frame(M=(mu[,c1]-mu[,c2]),A=0.5*(mu[,c1]+mu[,c2])))
        predictedVals=predict(help,data.frame(A=0.5*(mu[,c1]+mu[,c2])))
        plot(0.5*(mu[,c1]+mu[,c2]),mu[,c1]-mu[,c2]-predictedVals,main=paste("MAplot for mu after loess: cond ",c1, " against ",c2),xlab="A",ylab="M",pch=20)
        lines(c(0,15),c(0,0),col="grey")      
      }    
    }
  }
  cat("Plotted (A,M-loess) plots for \mu for all pairswise comparisons\n")
}




createMuSummaryTable<-function(){

  # Producing a table with summaries of the analysis.
  # The table has one row per gene and the following columns:
  #   row 1: the gene number (obtained when reading CEL files in R)
  #   row 2: an indicator for whether all parameters for the gene
  #          were monitored (inc S_gjcr's and H_gjcr).  1: if yes, 0: if no (default)
  #   rows 3 and 4: mean and sd of mu_1
  #   ...
  #   rows 2+2*C-1 and 2+2*C: mean and sd of mu_C (C: number conditions analysed)
  #   rows 2+2*C+1 to 2+2*C+8: mean(mu_2-mu_1), sd(mu_2-mu_1),
  #                             mean(mu_2-mu_1)/sd(mu_2-mu_1), rank(abs(mean(mu_2-mu_1)/sd(mu_2-mu_1))),         
  #                             mean(mu_2-mu_1-loess)/sd(mu_2-mu_1),rank(abs(mean(mu_2-mu_1-loess)/sd(mu_2-mu_1))),
  #                             loess, Posterior Probability of mu_2-mu_1-loess>0
  #   ...
  #   rows 2+2*C+numberPairs*8-7 to 2+2*C+numberPairs*8: as above but for the last pair.
  #
  # NB! 'loess': is the loess fitted curve to the MA plot of mean(mu_cond2) on mean(mu_cond1)
  # NB! 'last pair':  the pairs of conditions are ordered as: (1,2), (1,3), ... , (1,C)
  #                                                                  (2,3), ... , (2,C)
  #                                                                         ...
  #                                                                             (C-1,C)
  
  cat("Attempting to load 'bgxFunctionCall.Rdata'.\n")
  cat("If not successfull make sure your working directory is correct.\n")
  load("bgxFunctionCall.Rdata") 
  cat("  'bgxFunctionCall.Rdata' has successfully been loaded.\n")
 
  geneNumbersAnalysed=runParameterList$geneNumbersAnalysed
  geneNumbersWatched=runParameterList$geneNumbersWatched
  samples=sum(runParameterList$samplesets)
  probes=(runParameterList$numberProbesAnalysed)
  samplesets=runParameterList$samplesets
  
  if(is.null(geneNumbersWatched))
    watchedOrNot=rep(0,times=length(geneNumbersAnalysed))
  else
    watchedOrNot=replace(!is.na(match(geneNumbersAnalysed,geneNumbersWatched)),FALSE,0)

  meanMus=array(0,dim=c(length(geneNumbersAnalysed),length(samplesets)))
  sdMus=array(0,dim=c(length(geneNumbersAnalysed),length(samplesets)))

  numberPairs=(length(samplesets)*(length(samplesets)-1))/2
  meanDiffs=array(0,dim=c(length(geneNumbersAnalysed),numberPairs))
  sdDiffs=array(0,dim=c(length(geneNumbersAnalysed),numberPairs))
  postProbOfDiffIsPos=array(0,dim=c(length(geneNumbersAnalysed),numberPairs))
  loessCorr=array(0,dim=c(length(geneNumbersAnalysed),numberPairs))

  pairsSoFar=0
  for(y in 1:length(samplesets)){
   musy=matrix(scan(paste("mu.",y,sep="")),byrow=TRUE,nc=length(geneNumbersAnalysed))
   meanMus[,y]=apply(musy,2,mean)
   sdMus[,y]=apply(musy,2,sd)        
   if( y < length(samplesets)){
     for(x in (y+1):length(samplesets)){
       pairsSoFar=pairsSoFar+1    
       musx=matrix(scan(paste("mu.",x,sep="")),byrow=TRUE,nc=length(geneNumbersAnalysed))
       meanMus[,x]=apply(musx,2,mean)
       sdMus[,x]=apply(musx,2,sd)    
       help=loess(M~A,data.frame(M=(meanMus[,x]-meanMus[,y]),A=0.5*(meanMus[,x]+meanMus[,y])))       
       loessCorr[,pairsSoFar]=predict(help,data.frame(A=0.5*(meanMus[,x]+meanMus[,y])))       
       meanDiffs[,pairsSoFar]=apply(musx-musy,2,mean)
       sdDiffs[,pairsSoFar]=apply(musx-musy,2,sd)
       # posterior prob that difference is positive:
       loessarray<-matrix(data=rep(loessCorr[,pairsSoFar],times=dim(musx)[1]), nrow = 1024, ncol = length(geneNumbersAnalysed), byrow = TRUE, dimnames = NULL)
       posFreq <- function(x){
         return(length(x[x>0])/length(x))
       }
       postProbOfDiffIsPos[,pairsSoFar]=apply((musx-musy-loessarray),2,posFreq)
     }
   }
 } 

 # calculating loesscurve for MA plot of two sets of mean mu's:

 library(MASS)
 myMatrix =  matrix(0,nrow=length(geneNumbersAnalysed),ncol=(2+(2*length(samplesets))+(8*numberPairs)))
 myMatrix[,1]=geneNumbersAnalysed
 myMatrix[,2]=round(watchedOrNot,0)

 for(c in 1:length(samplesets)){
   myMatrix[,c*2+1]=meanMus[,c]
   myMatrix[,c*2+2]=sdMus[,c]
 }

 for(p in 1:numberPairs){
   myMatrix[,(length(samplesets)*2+2)+(p*8)-7]=meanDiffs[,p]
   myMatrix[,(length(samplesets)*2+2)+(p*8)-6]=sdDiffs[,p]
   myMatrix[,(length(samplesets)*2+2)+(p*8)-5]=meanDiffs[,p]/sdDiffs[,p]
   myMatrix[,(length(samplesets)*2+2)+(p*8)-4]=rank(abs(meanDiffs[,p]/sdDiffs[,p]))
   myMatrix[,(length(samplesets)*2+2)+(p*8)-3]=((meanDiffs[,p]-loessCorr[,p])/sdDiffs[,p])
   myMatrix[,(length(samplesets)*2+2)+(p*8)-2]=rank(abs((meanDiffs[,p]-loessCorr[,p])/sdDiffs[,p]))
   myMatrix[,(length(samplesets)*2+2)+(p*8)-1]=loessCorr[,p]
   myMatrix[,(length(samplesets)*2+2)+(p*8)]=postProbOfDiffIsPos[,p]
 }

 muheadings=c()
 for(c in 1:length(samplesets)){
   muheadings[2*c-1]=paste("E(mu",c,")",sep="")
   muheadings[2*c]=paste("SD(mu",c,")",sep="")
 } 
 diffheadings=c()
 pairsSoFar=0
 for(y in 1:length(samplesets)){
   if( (y+1) < length(samplesets)+1){
     for(x in (y+1):length(samplesets)){ 
       pairsSoFar=pairsSoFar+1     
       diffheadings[(pairsSoFar*8)-7]=paste("E(mu",y,"-mu",x,")",sep="")
       diffheadings[(pairsSoFar*8)-6]=paste("SD(mu",y,"-mu",x,")",sep="")
       diffheadings[(pairsSoFar*8)-5]=paste("E/SD",sep="")
       diffheadings[(pairsSoFar*8)-4]=paste("rank(|E/SD|)",sep="")
       diffheadings[(pairsSoFar*8)-3]=paste("(E-loess)/SD",sep="")
       diffheadings[(pairsSoFar*8)-2]=paste("rank(|(E-loess)/SD|)",sep="")
       diffheadings[(pairsSoFar*8)-1]=paste("loess")
       diffheadings[(pairsSoFar*8)]=paste("PPPD")
     }
   }
 }


 # rownames(myMatrix)=geneNames(data)[geneNumbersAnalysed]
 load("namesOfGenesAnalysed.Rdata")
 rownames(myMatrix)=namesOfGenesAnalysed

 colnames(myMatrix)=c(c("geneNumbers","monitoredOrNot"),muheadings,diffheadings)
 write.matrix(round(myMatrix,3),file="summaryTable.csv",sep=",")
 # write.table(list(geneNames=geneNames(data)[geneNumbersAnalysed]),file="geneNames.csv",row.names=FALSE)
 write.table(list(geneNames=namesOfGenesAnalysed),file="geneNames.csv",row.names=FALSE)
   cat("    Wrote file: summaryTable.csv\n",sep="")
   cat("    Wrote file: geneNames.csv\n",sep="")
}




makePostDistrMuBGXindex<-function(cond){

  # produces a table displaying quantiles of the posterior distribution
  # for the bgx gene expression index under condition 'cond'.
  # The table has the format: row(=gene) by column(=%-age)
  # Quantiles are: 2.5, 5, 10, 25, 33, 50, 66, 75, 90, 95, 97.5

  cat("Attempting to load 'bgxFunctionCall.Rdata'.\n")
  cat("If not successfull make sure your working directory is correct.\n")
  load("bgxFunctionCall.Rdata") 
  cat("    'bgxFunctionCall.Rdata' has successfully been loaded.\n")
 
  numberGenesAnalysed=length(runParameterList$geneNumbersAnalysed)
  numberConditions=length(runParameterList$samplesets)

  if(cond > numberConditions)  stop("You specified an invalid condition\n")

  mus=matrix(scan(paste("mu.",cond,sep="")),byrow=TRUE,nc=numberGenesAnalysed)
 
  myPercentiles<-function(x){
    # returns cut-off of the following %-iles: 2.5, 5, 10, 25, 33, 50, 66, 75, 90, 95, 97.5
    return(sort(x)[c(26,52,103,256,338,512,675,768,921,972,998)])
  }

  percentiles=apply(mus,2,myPercentiles)
  help=c(c("2.5%"), c("5%"), c("10%"), c("25%"), c("33%"), c("50%"), c("66%"), c("75%"), c("90%"), c("95%"), c("97.5%"))
  rownames(percentiles)=help
  write.matrix(round(t(percentiles),3),file=paste("postDistrBGXindexCond",cond,".csv",sep=""),sep=",")
  cat("    Wrote file: postDistrBGXindexCond",cond,".csv\n",sep="")
} 


makePostDistrMuBGXPairDiff<-function(cond1,cond2){

  # produces table displaying quantiles of the posterior distribution of 
  # differential expression between cond 1 and cond 2 (cond1-cond2).
  # Format: row(=gene) by column(diffExpr): BGXdiffTableCondYtoX
  # Quantiles given are: 2.5, 5, 10, 25, 33, 50, 66, 75, 90, 95, 97.5

  cat("Attempting to load 'bgxFunctionCall.Rdata'.\n")
  cat("If not successfull make sure your working directory is correct.\n")
  load("bgxFunctionCall.Rdata") 
  cat("  'bgxFunctionCall.Rdata' has successfully been loaded.\n")
 
  numberGenesAnalysed=length(runParameterList$geneNumbersAnalysed)
  numberConditions=length(runParameterList$samplesets)

  if(cond1 > numberConditions)  stop("You specified an invalid condition 1\n")
  if(cond2 > numberConditions)  stop("You specified an invalid condition 2\n")

  musX=matrix(scan(paste("mu.",cond1,sep="")),byrow=TRUE,nc=numberGenesAnalysed)
  musY=matrix(scan(paste("mu.",cond2,sep="")),byrow=TRUE,nc=numberGenesAnalysed) 

  diffs=musX-musY

  musXave=apply(musX,2,mean) 
  musYave=apply(musY,2,mean)
  
  help=loess(M~A,data.frame(M=(musXave-musYave),A=0.5*(musXave+musYave)))  
  predictedVals=predict(help,data.frame(A=0.5*(musXave+musYave)))
  loessArray=matrix(rep(predictedVals,times=15),nrow=numberGenesAnalysed,ncol=15,byrow=FALSE)

  myPercentiles<-function(x){
    # return(sort(x)[c(26,52,103,256,338,512,675,768,921,972,998)])
    return(sort(x)[c(26,52,103,154,205,256,338,512,675,768,819,870,921,972,998)])
  }
  
  # percentile values of diffs for gene g:  
  diffPercentiles=apply(diffs,2,myPercentiles)
  # help=c(c("2.5%"), c("5%"), c("10%"), c("25%"), c("33%"), c("50%"), c("66%"), c("75%"), c("90%"), c("95%"), c("97.5%"))
  help=c(c("2.5%"), c("5%"), c("10%"), c("15%"), c("20%"), c("25%"), c("33%"), c("50%"), c("66%"), c("75%"), c("80%"), c("85%"), c("90%"), c("95%"), c("97.5%"))
  
  rownames(diffPercentiles)=help
  write.matrix(round((t(diffPercentiles)-loessArray),3),file=paste("postDistrBGXdiffCond",cond1,"-",cond2,".csv",sep=""),sep=",")
  cat("    Wrote file: postDistrBGXdiffCond",cond1,"-",cond2,".csv\n",sep="")
} 




makeMuMAPlotCond2againstCond1<-function(condx,condy){

  # creates two plots:
  # 1: an MAplot of condy against condx and 2: an normalised MAplot of condy 
  # against condx (that is, the loess curve fitted to the original MAplkot is
  # subtracted.

  cat("Attempting to load 'bgxFunctionCall.Rdata'.\n")
  cat("If not successfull make sure your working directory is correct.\n")
  load("bgxFunctionCall.Rdata") 
  cat("  'bgxFunctionCall.Rdata' has successfully been loaded.\n")

  numberConditions=length(runParameterList$samplesets)
 
  if(condx > numberConditions)  stop("You specified an invalid condition 1\n")
  if(condy > numberConditions)  stop("You specified an invalid condition 2\n")

  dataTable=read.csv(file="summaryTable.csv",header = TRUE, sep = ",",dec=".")
  geneNames=read.csv(file="geneNames.csv",header = TRUE, sep = ",",dec=".")
  rownames(dataTable)=geneNames[[1]]
  numberGenesAnalysed=length(geneNames[[1]])
  par(mfrow=c(1,2)) 
  M=dataTable[,(condy*2)+1]-dataTable[,(condx*2)+1]
  A=0.5*(dataTable[,(condy*2)+1]+dataTable[,(condx*2)+1])
  plot(A,M,main="MA-plot before loess",xlab=paste("0.5*(cond(",condy,")+cond(",condx,"))"),ylab=paste("cond(",condy,")-cond(",condx,")"))
  lines(c(0,15),c(0,0),col=3)

  cat("  Plotted: (A,M) with M: condition",condy," - condition",condx,", and loess curve.\n",sep="")

  numberOfPair<-function(x,y,numberCond){
    theMin=min(x,y)
    theMax=max(x,y)
    if( theMin>1 ){
      help=0
      for(i in 2:max(2,theMin)){
        help=help+length(c(i:numberCond))
      }
      return(help+(theMax-theMin))
    }
    else{
      return(theMax-1)
    }
  }

  if( condy > condx ){
    points(0.5*(dataTable[,(condy*2)+1]+dataTable[,(condx*2)+1]),dataTable[,(numberConditions*2+2)+((numberOfPair(condx,condy,numberConditions))*8)-1],col=7,pch=20)
  }
  else{
    points(0.5*(dataTable[,(condy*2)+1]+dataTable[,(condx*2)+1]),-dataTable[,(numberConditions*2+2)+((numberOfPair(condx,condy,numberConditions))*8)-1],col=7,pch=20)
  } 
  if( condy > condx ){
    plot(A,M-dataTable[,(numberConditions*2+2)+((numberOfPair(condx,condy,numberConditions))*8)-1],main="MA-plot after loess",xlab=paste("0.5*(cond(",condy,")+cond(",condx,"))"),ylab=paste("cond(",condy,")-cond(",condx,")"))
  }
  else{
    plot(A,M+dataTable[,(numberConditions*2+2)+((numberOfPair(condx,condy,numberConditions))*8)-1],main="MA-plot after loess",xlab=paste("0.5*(cond(",condy,")+cond(",condx,"))"),ylab=paste("cond(",condy,")-cond(",condx,")"))
  }  
  lines(c(0,15),c(0,0),col=3)

  cat("  Plotted: (A,M-loess) with M: condition",condy," - condition",condx,"\n",sep="")
}



addkhighestRankedPointsToMuMAplot<-function(condx,condy,ranking,K,colour){

  # adds (condx,condy) muMAplot points of the k highest ranked genes
  # using 'ranking' as ranking criterion in colour 'colour' to the existing plot.
  # NB! A PLOT PRODUCED BY THE makeMuMAPlotCond2againstCond1 - function
  # SHOULD BE PRESENT IN THE PLOT DEVICE WHEN THIS FUNCTION IS CALLED.
  # THE IDEA IS THAT YOU CAN HIGHLIGHT GENES OF INTEREST. 

  cat("Attempting to load 'bgxFunctionCall.Rdata'.\n")
  cat("If not successfull make sure your working directory is correct.\n")
  load("bgxFunctionCall.Rdata") 
  cat("  'bgxFunctionCall.Rdata' has successfully been loaded.\n")
 
  numberGenesAnalysed=length(runParameterList$geneNumbersAnalysed)
  numberConditions=length(runParameterList$samplesets)

  cat("Adding points to MA plot for condition ",condy," against ",condx,":\n")

  if(condx > numberConditions)  stop("You specified an invalid condition 1\n")
  if(condy > numberConditions)  stop("You specified an invalid condition 2\n")
  if(K > numberGenesAnalysed)  stop("You specified too large k - there are not enough genes.\n")

  dataTable=read.csv(file="summaryTable.csv",header = TRUE, sep = ",",dec=".")
  M=dataTable[,(condy*2)+1]-dataTable[,(condx*2)+1]
  A=0.5*(dataTable[,(condy*2)+1]+dataTable[,(condx*2)+1])

  numberOfPair<-function(x,y,numberCond){
    theMin=min(x,y)
    theMax=max(x,y)
    if( theMin>1 ){
      help=0
      for(i in 2:max(2,theMin)){
        help=help+length(c(i:numberCond))
      }
      return(help+(theMax-theMin))
    }
    else{
      return(theMax-1)
    }
  }
  
  if( condy > condx){
    Mloess=M-dataTable[,(numberConditions*2+2)+((numberOfPair(condx,condy,numberConditions))*8)-1]
  }
  else{
    Mloess=M+dataTable[,(numberConditions*2+2)+((numberOfPair(condx,condy,numberConditions))*8)-1]
  }

  if(ranking=="ratio"){
    ranksForNormalisedData=dataTable[,(numberConditions*2+2)+(numberOfPair(condx,condy,numberConditions)*8)-2]
  }
  else if(ranking=="PPPD"){
    ranksForNormalisedData=dataTable[,(numberConditions*2+2)+(numberOfPair(condx,condy,numberConditions)*8)]
  }
  else{
    stop("third parameter ('ranking') not understood: you must use 'ratio' or 'PPPD'")
  }
  points(A[ranksForNormalisedData>(numberGenesAnalysed-K)],Mloess[ranksForNormalisedData>(numberGenesAnalysed-K)],col=paste(colour))

  cat("  added ",K," highest ranked points in colour=",paste(colour),"\n")  
  if(ranking=="ratio"){
    cat("(ranking based on abs((E-loess)/SD) values)\n") 
  }
  else{
    cat("(ranking based on PPPD (='posterior Probability of diff>0') values)\n") 
  }
}




plotCredibilityIntervalforDiff<-function(condx,condy,size,first,last,ranking){
 
  # plots stacked credibility intervals of the differences in expression
  # between condition 'condx' and 'condy' (mu(condx)-mu(condy)) 
  # of genes 'first' to 'last' when genes are ordered using 'ranking'.
  # The width of the credibility interval is given by 'size'.
  # Possible values are: 1,2,3,4,56 and 7 corresponding to:
  # 2.5-97.5, 5-95, 10-90, 15-85, 20-80, 25-75 and 33-67%. 
  # The numbers of the genes with credibility intervals for difference 
  # in expression that do no cover zero are denoted in
  # the right hand margin in red.

  cat("Attempting to load 'bgxFunctionCall.Rdata'.\n")
  cat("If not successfull make sure your working directory is correct.\n")
  load("bgxFunctionCall.Rdata") 
  cat("  'bgxFunctionCall.Rdata' has successfully been loaded.\n")
 
  numberGenesAnalysed=length(runParameterList$geneNumbersAnalysed)
  numberConditions=length(runParameterList$samplesets)

  if(condx > numberConditions)  stop("You specified an invalid condition 1\n")
  if(condy > numberConditions)  stop("You specified an invalid condition 2\n")

  # if( size == 1) cat("plotting credibility interval 2.5-97.5%\n")
  # else if( size == 2) cat("plotting credibility interval 5-95%\n")
  # else if( size == 3) cat("plotting credibility interval 10-90%\n")
  # else if( size == 4) cat("plotting credibility interval 25-75%\n")
  # else if( size == 5) cat("plotting credibility interval 33-67%\n")
  # else stop("Illegal value of third parameter 'size'. You must use 1,2,3,4 or 5\n")

  if( size == 1) cat("plotting credibility interval 2.5-97.5%\n")
  else if( size == 2) cat("plotting credibility interval 5-95%\n")
  else if( size == 3) cat("plotting credibility interval 10-90%\n")
  else if( size == 4) cat("plotting credibility interval 15-85%\n")
  else if( size == 5) cat("plotting credibility interval 20-80%\n")
  else if( size == 6) cat("plotting credibility interval 25-75%\n")
  else if( size == 7) cat("plotting credibility interval 33-67%\n")
  else stop("Illegal value of third parameter 'size'. You must use 1,2,3,4, 5,6 or 7\n")


  if( first > numberGenesAnalysed) stop("fourth parameter 'first' is too big. Must be smaller than numberGenesAnalysed: ",numberGenesAnalysed,"\n")
  if( last > numberGenesAnalysed) stop("fifth parameter 'last' is too big. Must be smaller than or equal to numberGenesAnalysed: ",numberGenesAnalysed,"\n")
  if( first > last ) stop("Error. Fourth parameter 'first' must be bigger than fifth parameter 'last'\n")

  summaryTable=read.csv(file="summaryTable.csv",header = TRUE, sep = ",",dec=".")
  credIntervalDiffTable=read.csv(file=paste("postDistrBGXdiffCond",condx,"-",condy,".csv",sep=""))
  
  numberOfPair<-function(x,y,numberCond){
    theMin=min(x,y)
    theMax=max(x,y)
    if( theMin>1 ){
      help=0
      for(i in 2:max(2,theMin)){
        help=help+length(c(i:numberCond))
      }
      return(help+(theMax-theMin))
    }
    else{
      return(theMax-1)
    }
  }

  if(ranking=="ratio"){
    values=summaryTable[,(numberConditions*2+2)+(numberOfPair(condx,condy,numberConditions)*8)-3]
    reverseOrder=order(-abs(values))
    cat("ranking on abs((E-loess)/SD) values\n")
  }
  else if(ranking=="PPPD"){
    values=summaryTable[,(numberConditions*2+2)+(numberOfPair(condx,condy,numberConditions)*8)]
    reverseOrder=order(-abs(values))
    cat("ranking on PPPD values\n")
  }
  else{
    stop("sixth parameter ('ranking') not understood: you must use 'ratio' or 'PPPD'")
  }
 
  fromgene=first
  togene=last+1
  leftPerc=size
  # rightPerc=12-size
  rightPerc=16-size

  plot(1,1,xlim=range(-5,5),ylim=range(fromgene,togene),type="n",main=paste("Credibility interval size: ",size,sep=""),xlab="difference in expression",ylab=paste("genes ordered by: ",ranking,sep=""))
  lines(c(0,0),c(fromgene-1,togene+1),col=3)
  for(g in fromgene:(togene-1)){
    lines(credIntervalDiffTable[reverseOrder[g],c(leftPerc,rightPerc)],c(g,g))
    if( credIntervalDiffTable[reverseOrder[g],leftPerc] > 0 )
      mtext(c(1:numberGenesAnalysed)[reverseOrder[g]],side=4,at=g,cex=0.6,las=2,col=2) 
    else if( credIntervalDiffTable[reverseOrder[g],rightPerc] < 0 )
      mtext(c(1:numberGenesAnalysed)[reverseOrder[g]],side=4,at=g,cex=0.6,las=2,col=2)   
  }
}

