/*
 *  This file is part of BGX, the Bayesian Gene eXpression program.
 *  Copyright (c) 2003-2004  Graeme Ambler <graeme@ambler.me.uk>
 *                2004       Anne-Mette Hein <a.hein@imperial.ac.uk>
 *
 *  BGX is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License, version 2, as
 *  published by the Free Software Foundation.
 *
 *  BGX is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include "rand.hh"
#include "rundir.hh"
#include "TierQuan.hh"
#include "qnorm.h"
#include "sokal.hh"
#include <string>
#include <iostream>
#include <fstream>
#include <valarray>
#include <vector>

// Some convenient definitions
typedef std::valarray<double> array;
typedef std::valarray<std::valarray<double> > array2d;

void stringcpy(char* p, const std::string& s)
{
  s.copy(p,std::string::npos);
  p[s.length()] = 0; // Add C-style string terminator
}

// This stuff is in a separate file for convenience of editting, not 
// because we want to compile it separately or anything.
#include "bgx_updates.hh"

inline double TN_quantile(double p, double mu, double sigma)
{
  return mu-sigma*qnorm(pnorm(mu/sigma)*(1-p));
}

#include "bgx.hh"

void bgx(double* pm, double* mm, int* samples, int* conditions, 
	 int* probes, int* genes, int* numberGenesToWatch, int* samplesets, int* probesets, 
       int* whichGenesToWatch, int* whichProbesToWatch, 
	 int* iter, int* burnin, double* s_jmp, double* h_jmp, 
	 double* mu_jmp, double* tau_jmp, double* lambda_jmp, 
	 double* eta_jmp, double* quan, int* num_quan, int* init_size, 
	 int* subsample, int* output, char** dirname, int* seed)
{
  Random rand(*seed);

  // Create output file and directory
  std::string filename;
  std::string run_dir = rundir("run");

  // Write out a summary of the call
  filename=run_dir+"/summary.txt";
  std::ofstream summary(filename.c_str());
  summary << "The program was called with the following parameter values:"
	  << std::endl
	  << std::endl
	  << "Number of samples:\t" << *samples << std::endl
	  << "Number of conditions:\t" << *conditions << std::endl
	  << "Number of probes:\t" << *probes << std::endl
	  << "Number of genes:\t" << *genes << std::endl << std::endl
        << "Number of genes to monitor fully:\t" << *numberGenesToWatch << std::endl 
    	  << "Number of post-burn-in sweeps:\t" << *iter << std::endl
	  << "Subsampling interval:\t\t" << *subsample << std::endl
	  << "Number of burn-in sweeps:\t" << *burnin << std::endl
	  << std::endl
	  << "Spread of jumps in S:\t\t" << *s_jmp << std::endl
	  << "Spread of jumps in H:\t\t" << *h_jmp << std::endl
	  << "Spread of jumps in mu:\t\t" << *mu_jmp << std::endl
	  << "Spread of jumps in tau:\t\t" << *tau_jmp << std::endl
	  << "Spread of jumps in lambda:\t" << *lambda_jmp << std::endl
	  << "Spread of jumps in eta:\t\t" << *eta_jmp << std::endl
	  << std::endl
	  << "Name of output directory:\t" << run_dir << std::endl
	  << std::endl
	  << "Seed for random number generator:\t" << *seed << std::endl
	  << std::endl
	  << "CEL files analysed: " << std::endl
	  << std::endl;
  summary.close();

  // Write out a progress file:
  filename=run_dir+"/progress";
  std::ofstream progress(filename.c_str());
  progress << "Progress monitoring started: " << std::endl;
  progress.close();

  // Return the name of the ouput directory
  stringcpy(*dirname,run_dir);

  filename=run_dir+"/progress";
  progress.open(filename.c_str(),std::ios::app);
  progress << "  Initialising variables: " << std::endl;
  progress.close();

  array2d medians(*conditions);
  for(int j=0; j<*conditions; ++j){
    medians[j].resize(*genes);
    for(int i=0; i<*genes; ++i)
      medians[j][i]=0;
  }

  double* thetagave = new double[*genes**conditions];
  for(int i=0; i<*genes**conditions; ++i)
    thetagave[i]=0;
  filename=run_dir+"/thetagave";
  std::ofstream thetagave_(filename.c_str());

  // Calculate quantiles of the medians.
  std::vector<Quantiles_T> quantiles(*genes**conditions);
  std::vector<double> p(*num_quan);
  for(int i=0; i<*num_quan; ++i) p[i]=quan[i];
  for(int i=0; i<*genes**conditions; ++i)
    quantiles[i].Initialize(p,*init_size);
  // Open a file to print the quantiles to
  filename=run_dir+"/quantiles";
  std::ofstream Quantiles(filename.c_str());

  filename=run_dir+"/tau2";
  std::ofstream tau2(filename.c_str());
  filename=run_dir+"/phi";
  std::ofstream phi_(filename.c_str());
  filename=run_dir+"/lambda";
  std::ofstream lambda_(filename.c_str());
  filename=run_dir+"/eta2";
  std::ofstream eta2(filename.c_str());

  // BEGIN TRACE /////////////////////////////////////////
  // Declare (but don't use) the output files for the median trace.
  // These are only initialised if we are doing the traces.
  std::ofstream *MT = new std::ofstream[*conditions];
  std::ofstream *MUS = new std::ofstream[*conditions];
  std::ofstream *SIGMA2S = new std::ofstream[*conditions];
  // Now initialise them if they are used
  if(*output>0)
    {
      std::string num;
      for(int j=0; j<*conditions; ++j){
	int_to_string(j+1,num);
	filename=run_dir+"/theta."+num;
	MT[j].open(filename.c_str());
      filename=run_dir+"/mu."+num;
      MUS[j].open(filename.c_str());
      filename=run_dir+"/sigma2."+num;
      SIGMA2S[j].open(filename.c_str());
      }
    }
  // END TRACE //////////////////////////////////////////

  // BEGIN FULL /////////////////////////////////////////
  // Declare (but don't use) the output files for the x and y trace.
  // These are only initialised if we are doing the full output.
   std::ofstream *XS = new std::ofstream[*samples];
   std::ofstream *YS = new std::ofstream[*samples];
  // Now initialise them if they are used
  if(*numberGenesToWatch>0)
    {
      std::string num;
      for(int j=0; j<*samples; ++j){
	int_to_string(j+1,num);
	filename=run_dir+"/x."+num;
	XS[j].open(filename.c_str());
      filename=run_dir+"/y."+num;
      YS[j].open(filename.c_str());
     }
   }
  // END FULL  //////////////////////////////

  // BEGIN MC ERRORS
  double** MCErrors=0;
  std::ofstream MCE;
  // If we are calculating the Monte-Carlo errors initialise the array and 
  // the output file
  if(*output>3)
    {
      MCErrors = new double*[*genes**conditions];
      for(int i=0; i<*genes**conditions; ++i)
	MCErrors[i] = new double[*iter];
      filename=run_dir+"/MCE";
      MCE.open(filename.c_str());
    }
  // END MC ERRORS //////////////////////////////////////////////
 
  // BEGIN DIAGNOSTICS
  // Declare (but don't use) the variables for the diagnostics.
  // These are only initialised if we are doing the diagnostics.
  double* xave=0;
  double* yave=0;
  double* mugave=0;
  double* sigmag2ave=0;
  double* sacc=0;
  double* hacc=0;
  double* mugacc=0;
  double* taugacc=0;
  double* lambdaacc=0;
  double* etaacc=0;
  std::ofstream a_;
  std::ofstream b2;
  std::ofstream beta_;
  std::ofstream mugave_;
  std::ofstream sigmag2ave_;
  std::ofstream xave_;
  std::ofstream yave_;
  std::ofstream sacc_;
  std::ofstream hacc_;
  std::ofstream mugacc_;
  std::ofstream taugacc_;
  std::ofstream lambdaacc_;
  std::ofstream etaacc_;

  std::vector<Quantiles_T> x_med;
  std::vector<Quantiles_T> y_med;
  std::vector<Quantiles_T> mug_med;
  std::vector<Quantiles_T> sigmag2_med;
  std::ofstream X_Med;
  std::ofstream Y_Med;
  std::ofstream Mug_Med;
  std::ofstream Sigmag2_Med;

  // If we are doing the diagnostics, initialise all the diagnostic variables 
  // we have declared.
  if(*output>2)
    {
      // We now have to allocate space for and initialize the arrays
      xave = new double[*probes**samples];
      yave = new double[*probes**samples];
      mugave = new double[*genes**conditions];
      sigmag2ave = new double[*genes**conditions];
      sacc = new double[*probes**samples];
      hacc = new double[*probes**samples];
      mugacc = new double[*genes**conditions];
      taugacc = new double[*genes**conditions];
      lambdaacc = new double[*samples];
      etaacc = new double[*samples];
      for(int i=0; i<*probes**samples; ++i)
	xave[i]=yave[i]=sacc[i]=hacc[i]=0;
      for(int i=0; i<*genes**conditions; ++i)
	mugave[i]=sigmag2ave[i]=mugacc[i]=taugacc[i]=0;
      for(int i=0; i<*samples; ++i)
	lambdaacc[i]=etaacc[i]=0;

      // Open output files
      filename=run_dir+"/a";
      a_.open(filename.c_str());
      filename=run_dir+"/b2";
      b2.open(filename.c_str());
      filename=run_dir+"/beta";
      beta_.open(filename.c_str());
      filename=run_dir+"/mugave";
      mugave_.open(filename.c_str());
      filename=run_dir+"/sigmag2ave";
      sigmag2ave_.open(filename.c_str());
      filename=run_dir+"/xave";
      xave_.open(filename.c_str());
      filename=run_dir+"/yave";
      yave_.open(filename.c_str());
      filename=run_dir+"/sacc";
      sacc_.open(filename.c_str());
      filename=run_dir+"/hacc";
      hacc_.open(filename.c_str());
      filename=run_dir+"/mugacc";
      mugacc_.open(filename.c_str());
      filename=run_dir+"/taugacc";
      taugacc_.open(filename.c_str());
      filename=run_dir+"/lambdaacc";
      lambdaacc_.open(filename.c_str());
      filename=run_dir+"/etaacc";
      etaacc_.open(filename.c_str());

      // Use Tierney's algorithm to give approximate medians of s,h,mu,sigma2.
      std::vector<double> half(1);
      half[0]=0.5;

      x_med.resize(*probes**samples);
      y_med.resize(*probes**samples);

      for(int i=0; i<*probes**samples; ++i){
    	  x_med[i].Initialize(half,*init_size);
	  y_med[i].Initialize(half,*init_size);   
      }

      mug_med.resize(*genes**conditions);
      sigmag2_med.resize(*genes**conditions);
      for(int i=0; i<*genes**conditions; ++i){
	mug_med[i].Initialize(half,*init_size);
	sigmag2_med[i].Initialize(half,*init_size);
      }
      filename=run_dir+"/x_med";
      X_Med.open(filename.c_str());
      filename=run_dir+"/y_med";
      Y_Med.open(filename.c_str());
      filename=run_dir+"/mug_med";
      Mug_Med.open(filename.c_str());
      filename=run_dir+"/sigmag2_med";
      Sigmag2_Med.open(filename.c_str());
    }
  // END DIAGNOSTICS ///////////////////////////////////////
 
  filename=run_dir+"/progress";
  progress.open(filename.c_str(),std::ios::app);
  progress << "  initialising ended." << std::endl;
  progress.close();

  filename=run_dir+"/progress";
  progress.open(filename.c_str(),std::ios::app);
  progress << "  Reading probe level data into C++style arrays:" << std::endl;
  progress.close();

  // Read data into C++-style arrays for ease of manipulation
  array2d PM(*samples);
  array2d MM(*samples);
  for(int j=0; j<*samples; ++j){
    PM[j].resize(*probes);
    MM[j].resize(*probes);
    for(int i=0; i<*probes; ++i){
      PM[j][i]=*(pm+i+*probes*j);
      MM[j][i]=*(mm+i+*probes*j);
    }
  }

  filename=run_dir+"/progress";
  progress.open(filename.c_str(),std::ios::app);
  progress << "  data read in." << std::endl;
  progress.close();

  filename=run_dir+"/progress";
  progress.open(filename.c_str(),std::ios::app);
  progress << "  Setting initial parameter values:" << std::endl;
  progress.close();

  // Define parameters and pick some vaguely sensible values for 
  // the initial values.
  double phi=0.18;
  array2d s(*samples);
  array2d h(*samples);
  array tau(*samples);
  array beta(*samples);
  array lambda(*samples);
  // Throughout the code we use eta to refer to eta^{-2}.
  array eta(*samples);
  for(int j=0; j<*samples; ++j){
    s[j].resize(*probes);
    h[j].resize(*probes);
    tau[j]=0.0005;
    beta[j]=0;
    double mean_lh=0; double var_lh=0;
    const double tmp2=log(1.1);
    for(int i=0; i<*probes; ++i){
      double tmp = (PM[j][i]-MM[j][i])/(1-phi);
      s[j][i] = 0.1>tmp ? 0.1 : tmp;
      tmp = (phi*PM[j][i]-MM[j][i])/(phi-1);
      if(0.1>tmp){
	h[j][i]=0.1;
	mean_lh += tmp2;
      }
      else{
	h[j][i]=tmp;
	mean_lh+=log(tmp+1);
      }
    }
    mean_lh /= *probes;
    for(int i=0; i<*probes; ++i){
      double L = log(h[j][i]+1);
      var_lh += (L-mean_lh)*(L-mean_lh);
    }
    var_lh /= *probes-1;
    lambda[j]=mean_lh;
    eta[j]=1.0/var_lh;
  }

  // AMs addition: Obtaining Empirical Bayes like
  // estiamtes of a and b^2:

  array2d ebS(*samples);

  // setting empirical Bayes values of signals: ebS
  std::vector<std::pair<int,int> > badSets; // probe sets which have PM<=MM for all probes
  for(int j=0; j<*samples; ++j){ 
     ebS[j].resize(*probes);  
     int probeCounter = 0;
     for(int i=0; i<*genes; ++i){        
        // If PM>MM: ebS = PM-MM. Furthermore, for each probe set: 
        // find minimum positive diff and number probe pairs with PM>MM. 
        int positive = 0;
	  // We test minimumDiff == HUGE_VAL below so be careful changing this!
        double minimumDiff = HUGE_VAL;
        for(int p=0; p<probesets[i]; ++p){
           if ( PM[j][probeCounter+p] > MM[j][probeCounter+p] ){
              ebS[j][probeCounter+p] = PM[j][probeCounter+p] - MM[j][probeCounter+p];
              positive++;
              if( minimumDiff > ebS[j][probeCounter+p] ){
                 minimumDiff = ebS[j][probeCounter+p];
              }
           }
        }
	  if(minimumDiff == HUGE_VAL){ // PM<=MM for all probes in probeset
	     std::pair<int,int> tempPair(j,i);
	     badSets.push_back(tempPair);
	  }
        for(int p=0; p<probesets[i]; ++p){
           // If PM<MM: ebS = fraction of minimum positive PM-MM
           if ( PM[j][probeCounter+p] < MM[j][probeCounter+p] ){    
              ebS[j][probeCounter+p] = (double(positive)/double(probesets[i]))*minimumDiff;
           }
           // if (!(PM<MM)) and (!(PM>MM)) (that is, if PM=MM): ebS=0
           else if( !(PM[j][probeCounter+p] > MM[j][probeCounter+p]) ){
              ebS[j][probeCounter+p] = 0.0;
           }
        }
        probeCounter += probesets[i];
     }
  }

  // For each (g,c), calculate variance over set consisting of
  // the values log(ebS_{g,j,c,r}+1), j=1,...,J_g, r=1,...,R_c

  array2d vars(*conditions);
  array a(*conditions);
  // Throughout the code we use b to refer to b^{-2}. 
  array b(*conditions);
  
  bool badData = false;
  
  int sampleCounter = 0;
  for(int j=0; j<*conditions; ++j){
     vars[j].resize(*genes);
     int probeCounter = 0;
     for(int i=0; i<*genes; ++i){
        double tempS = 0;
        double tempSS = 0;                
        
        for(int sam=0; sam<samplesets[j]; ++sam){
           for(int p=0; p<probesets[i]; ++p){
	      double logEbS=log(ebS[sampleCounter+sam][probeCounter+p]+1.0);
              tempS += logEbS;
              tempSS += logEbS*logEbS;
           }
        }
        probeCounter += probesets[i];
        vars[j][i] = (1.0/(double(probesets[i] * samplesets[j])-1.0))*
                       (tempSS-(tempS*tempS/double(probesets[i]*samplesets[j])));
     }
     sampleCounter += samplesets[j];

     // For each condition, calculate the mean and the variance
     // (or precision) of the log-variances log(vars[j][i]):
     double tempS = 0.0;
     double tempSS = 0.0;
     double minVar = HUGE_VAL;
     for(int i=0; i<*genes; ++i)
       if(vars[j][i]<minVar && vars[j][i]>0) minVar=vars[j][i];
     if(minVar == HUGE_VAL){ // PM<=MM for all probe sets.
       // Set a,b to default values.
       badData = true;
       a[j] = 0;
       b[j] = 3;
     }
     else{
       minVar=log(minVar);
       for(int i=0; i<*genes; ++i){
         double logVar = 0.0;
         if(vars[j][i]>0.0) logVar = log(vars[j][i]);
         else logVar = minVar;
         tempS += logVar;
         tempSS += logVar*logVar;
       }
       a[j] = tempS / double(*genes);
       // NB! b[j] is the precision (=1/variance): b^{-2}
       b[j] = (double(*genes)-1)/(tempSS-(tempS*tempS / double(*genes)));
     }
     std::cout << "empirical Bayes-like estimates of a[" << j << "]: " << a[j]
               << "  b[" << j << "]: " << b[j] << std::endl;
  }
  
  filename=run_dir+"/empBayesEst.txt";
  std::ofstream empBayesEst_(filename.c_str());
  empBayesEst_ << "Empirical Bayes like estimates of a and b^{-2} for each condition:" 
               << std::endl << std::endl;
  empBayesEst_ << '\t' << "condition"  <<  '\t' << "a"  <<  '\t' << "b^{-2}"  << std::endl;
  for(int j=0; j<*conditions; j++){
     empBayesEst_ << '\t'  << j << '\t'  <<  a[j] << '\t'  <<  b[j] << std::endl;
  }
  if(badSets.size()>1){
    std::cout << "Warning:  There were " << badSets.size() 
              << " probe sets with PM<=MM for all probes." << std::endl;
    
    empBayesEst_ << std::endl << "There were " << badSets.size() 
                 << " probe sets with PM<=MM for all probes." << std::endl;
  }
  else if(badSets.size()==1){
    std::cout << "Warning:  There were " << badSets.size() 
              << " probe sets with PM<=MM for all probes." << std::endl;

    empBayesEst_ << std::endl << "There were " << badSets.size() 
                 << " probe set with PM<=MM for all probes." << std::endl;

  }
  empBayesEst_.close();
             
  // AMs addition end!

  // Setting initial values for mu_gc and tau^2_gc

  array2d mu(*conditions);
  array2d taug(*conditions);

  for(int j=0; j<*conditions; ++j){
    mu[j].resize(*genes);
    taug[j].resize(*genes);
  }

  if(badData){ // PM<=MM for all probe sets.

    for(int i=0; i<*genes; ++i){
      for(int j=0; j<*conditions; ++j){
        mu[j][i] = 0;
        taug[j][i] = 1;
      }
    }
  }
  else{
        
    int p_loc=0;
    double minVar = HUGE_VAL;

    for(int i=0; i<*genes; ++i){
      int s_loc=0;

      for(int j=0; j<*conditions; ++j){
        double mean_ls=0; double var_ls=0;
        for(int sam=0; sam<samplesets[j]; ++sam){
	  for(int p=0; p<probesets[i]; ++p)
	    mean_ls += log(s[sam+s_loc][p+p_loc]+1);
        }
        mean_ls /= samplesets[j]*probesets[i];
        for(int sam=0; sam<samplesets[j]; ++sam){
	  for(int p=0; p<probesets[i]; ++p){
	    double L = log(s[sam+s_loc][p+p_loc]+1);
	    var_ls += (L-mean_ls)*(L-mean_ls);
	  }
        }
        var_ls /= samplesets[j]*probesets[i]-1;
        mu[j][i]=mean_ls;
        if(var_ls>0.0){
          taug[j][i]=1/var_ls;
	    if(var_ls<minVar) minVar=var_ls;
        }
        else{ // AMs addititon begin
          taug[j][i]=1/minVar;
        }  // AMs addition end
        s_loc += samplesets[j];
      }
      p_loc += probesets[i];
    }
  }

  filename=run_dir+"/progress";
  progress.open(filename.c_str(),std::ios::app);
  progress << "  Initial values have been set." << std::endl;
  progress.close();

  // Define the update objects.  For random walk Metropolis (RWM) steps, this 
  // involves creating an acceptance probability functor (e.g. S_T AccS(...)) 
  // and a RWM object (e.g. RWM<S_T,array2d> S(...)).  For Gibbs moves we just
  // create a single update object (e.g. A_T A(...)).
  // These are all defined in bgx_updates.hh.

  S_T AccS(PM,MM,h,phi,mu,taug,tau,beta,probesets,samplesets);
  RWM<S_T,array2d> S(s,AccS,*s_jmp,&rand);

  H_T AccH(PM,MM,s,phi,lambda,eta,tau,beta);
  RWM<H_T,array2d> H(h,AccH,*h_jmp,&rand);

  Mu_T AccMu(s,taug,probesets,samplesets);
  RWM<Mu_T,array2d> Mu(mu,AccMu,*mu_jmp,&rand);

  TauG_T AccTauG(s,mu,a,b,probesets,samplesets);
  RWM<TauG_T,array2d> TauG(taug,AccTauG,*tau_jmp,&rand);

  Lambda_T AccLambda(h,eta,0,0.001);
  RWM<Lambda_T> Lambda(lambda,AccLambda,*lambda_jmp,&rand);

  Eta_T AccEta(h,lambda,0.001,0.001);
  RWM<Eta_T> Eta(eta,AccEta,*eta_jmp,&rand);

  // Not updating a and b due to slow mixing
  //    A_T A(a,taug,b,p_zero,0.0625,&rand);
  //    B_T B(b,taug,a,1.2,0.3,&rand);

  Phi_T Phi(phi,MM,s,h,tau,beta,&rand);

  Tau_T Tau(tau,PM,MM,s,h,phi,beta,0.001,0.001,&rand);

  // Not updating Beta due to computational problems
  //    Beta_T Beta(beta,PM,MM,s,h,phi,tau,0,10,&rand);

  filename=run_dir+"/progress";
  progress.open(filename.c_str(),std::ios::app);
  progress << "Starting burn-in phase:" << std::endl;
  progress.close();

  // Run the Markov chain: Burn-in phase
  for(int sweep=0; sweep<*burnin; ++sweep){
    S.Update();
    H.Update();

    Mu.Update();
    TauG.Update();

    Lambda.Update();
    Eta.Update();

    Tau.Update();
    Phi.Update();

    // Not updating a and b due to slow mixing
    //    A.Update();
    //    B.Update();

    // Not updating Beta due to computational problems
    //    Beta.Update();
  }

  filename=run_dir+"/progress";
  progress.open(filename.c_str(),std::ios::app);
  progress << "  Burn-in phase ended." << std::endl;
  progress.close();

  S.Reset();
  H.Reset();

  Mu.Reset();
  TauG.Reset();

  Lambda.Reset();
  Eta.Reset();

  // start the sampling:

  filename=run_dir+"/progress";
  progress.open(filename.c_str(),std::ios::app);
  progress << "Start to sample. Number sweeps performed so far:" << std::endl;
  progress.close();

  for(int sweep=0; sweep<*iter; ++sweep){
    S.Update();
    H.Update();

    Mu.Update();
    TauG.Update();

    Lambda.Update();
    Eta.Update();

    Tau.Update();
    Phi.Update();

    // Not updating a and b due to slow mixing
    //    A.Update();
    //    B.Update();

    // Not updating Beta due to identifiability problems
    //    Beta.Update();

    if(sweep>=0){
      // Store medians and update means
      for(int c=0; c<*conditions; ++c)
	for(int g=0; g<*genes; ++g){
	  medians[c][g] = TN_quantile(0.5,mu[c][g],sqrt(1/taug[c][g]));
	  thetagave[g+*genes*c] += medians[c][g];
	  // BEGIN MC ERRORS ////////////////////
	  if(*output>3)
	    MCErrors[g+*genes*c][sweep]=medians[c][g];
	  // END MC ERRORS //////////////////////
	}

      // BEGIN DIAGNOSTICS  ///////////////////
      if(*output>2)
	{
	  for(int sample=0; sample<*samples; ++sample){
	    // Update means of S and H
	    for(int p=0; p<*probes; ++p){
	      xave[p+*probes*sample] += log(s[sample][p]+1);
	      yave[p+*probes*sample] += log(h[sample][p]+1);
	    }
	  }
	  for(int c=0; c<*conditions; ++c){
	    // Update averages of the mu's and tau's
	    for(int g=0; g<*genes; ++g){
	      mugave[g+*genes*c] += mu[c][g];
	      sigmag2ave[g+*genes*c] += 1/taug[c][g];
	    }
	  }
	}
      // END DIAGNOSTICS  //////////////////////////

      if(!(sweep%*subsample)){
	// Update quantiles of medians
	for(int c=0; c<*conditions; ++c)
	  for(int g=0; g<*genes; ++g){
	    quantiles[g+*genes*c](medians[c][g]);
	  }
	phi_ << phi << '\n';
	for(int sample=0; sample<*samples; ++sample){
	  tau2 << 1/tau[sample] << '\t';
	  lambda_ << lambda[sample] << '\t';
	  eta2 << 1/eta[sample] << '\t';
	}
	tau2 << '\n';
	lambda_ << '\n';
	eta2 << '\n';

	// BEGIN TRACE //////////////////////////////////////
	if(*output>0)
	  {
          for(int c=0; c<*conditions; ++c){
	      // Output medians
	      for(int g=0; g<*genes; ++g){
              MT[c] << medians[c][g] << '\t';
	        MUS[c] << mu[c][g] << '\t';
              SIGMA2S[c] << 1.0/taug[c][g] << '\t';
            }
	      MT[c] << '\n';
	      MUS[c] << '\n';
            SIGMA2S[c] << '\n';
	    }
	  }
	// END TRACE  ////////////////////////////////////////
      
      // BEGIN FULL //////////////////////////////////////
	if(*numberGenesToWatch>0)
	  {
          for(int sample=0; sample<*samples; ++sample){
            // output x and y
            for(int g=0; g<*numberGenesToWatch; g++){
              for(int p=0; p<probesets[whichGenesToWatch[g]]; p++){
                XS[sample] << log(s[sample][(whichProbesToWatch[g]-1)+p]+1) << '\t';
                YS[sample] << log(h[sample][(whichProbesToWatch[g]-1)+p]+1) << '\t';
              }
            }
            XS[sample] << '\n';
            YS[sample] << '\n';
          }
    	  }
	// END FULL  ////////////////////////////////////////

	// BEGIN DIAGNOSTICS  ////////////////////////////////
	if(*output>2)
	  {
          for(int sample=0; sample<*samples; ++sample){
	      beta_ << beta[sample] << '\t';
	      // Update medians of S and H
	      for(int p=0; p<*probes; ++p){
		x_med[p+*probes*sample](log(s[sample][p]+1));
		y_med[p+*probes*sample](log(h[sample][p]+1));
	      }
	    }
	    beta_ << '\n';
	    for(int c=0; c<*conditions; ++c){
	      a_ << a[c] << '\t';
	      b2 << b[c] << '\t';
	      // Update medians of the mu's and tau's
	      for(int g=0; g<*genes; ++g){
		mug_med[g+*genes*c](mu[c][g]);
		sigmag2_med[g+*genes*c](1/taug[c][g]);
	      }
	    }
	    a_ << '\n';
	    b2 << '\n';
	  }
	// END DIAGNOSTICS ////////////////////////////////

      filename=run_dir+"/progress";
      progress.open(filename.c_str(),std::ios::app);
      progress << sweep << std::endl;
      progress.close();

      }
    }
  }

  filename=run_dir+"/progress";
  progress.open(filename.c_str(),std::ios::app);
  progress << "Final sweep performed." << std::endl;
  progress.close();

  // BEGIN MC ERRORS ////////////////////////////////////
  // Must do this before we delete thetagave!
  if(*output>3)
    {
      for(int i=0; i<*genes**conditions; ++i)
	{
	  double var=0;
	  double tau=0;
	  int m=0;
	  if(sokal(*iter,MCErrors[i],&var,&tau,&m)!=0) MCE << (*iter) << std::endl;
	  else MCE << tau << '\t' << var << '\t' << tau*var/(*iter) << std::endl;
	  delete[] MCErrors[i];
	}
      delete[] MCErrors;
      MCE.close();
    }
	
  // END MC ERRORS ////////////////////////////////////

  for(int c=0; c<*conditions; ++c){
    for(int g=0; g<*genes; ++g){
      thetagave_ << thetagave[g+*genes*c] /(*iter) << '\t';
      std::vector<double> q=quantiles[g+*genes*c].Output();
      for(int i=0; i<*num_quan; ++i) Quantiles << q[i] << '\t';
      Quantiles << '\n';
    }
    thetagave_ << '\n';
  }
  delete[] thetagave;
  thetagave_.close();
  Quantiles.close();
  tau2.close();
  phi_.close();
  lambda_.close();
  eta2.close();

  // BEGIN TRACE /////////////////////////////////////
  if(*output>0)
    {
      for(int j=0; j<*conditions; ++j){
        MT[j].close();
        MUS[j].close();
        SIGMA2S[j].close();
      }
    }
  // END TRACE ////////////////////////////////////////

  // BEGIN FULL /////////////////////////////////////
  if(*numberGenesToWatch>0)
    {
      for(int sample=0; sample<*conditions; ++sample){
        XS[sample].close();
        YS[sample].close();
      }
    }
  // END FULL ////////////////////////////////////////

  // BEGIN DIAGNOSTICS ////////////////////////////////
  if(*output>2)
    {
      // Calculate acceptance probabilities
      S.pAccept(*iter,sacc);
      H.pAccept(*iter,hacc);
      Mu.pAccept(*iter,mugacc);
      TauG.pAccept(*iter,taugacc);
      Lambda.pAccept(*iter,lambdaacc);
      Eta.pAccept(*iter,etaacc);

      for(int c=0; c<*conditions; ++c){
	for(int g=0; g<*genes; ++g){
	  mugave_ << mugave[g+*genes*c]/(*iter) << '\t';
	  std::vector<double> med_tmp=mug_med[g+*genes*c].Output();
	  Mug_Med << med_tmp[0] << '\t';
	  sigmag2ave_ << sigmag2ave[g+*genes*c] /(*iter) << '\t';
	  med_tmp=sigmag2_med[g+*genes*c].Output();
	  Sigmag2_Med << med_tmp[0] << '\t';
	  mugacc_ << mugacc[g+*genes*c] << '\t';
	  taugacc_ << taugacc[g+*genes*c] << '\t';
	}
	mugave_ << '\n';
	Mug_Med << '\n';
	sigmag2ave_ << '\n';
	Sigmag2_Med << '\n';
	mugacc_ << '\n';
	taugacc_ << '\n';
      }
  
      for(int sample=0; sample<*samples; ++sample){
	for(int p=0; p<*probes; ++p){
	  xave_ << xave[p+*probes*sample]/(*iter) << '\t';
	  yave_ << yave[p+*probes*sample]/(*iter) << '\t';
	  std::vector<double> med_tmp=x_med[p+*probes*sample].Output();
	  X_Med << med_tmp[0] << '\t';
	  med_tmp=y_med[p+*probes*sample].Output();
	  Y_Med << med_tmp[0] << '\t';
	  sacc_ << sacc[p+*probes*sample] << '\t';
	  hacc_ << hacc[p+*probes*sample] << '\t';
	}
	xave_ << '\n';
	yave_ << '\n';
	X_Med << '\n';
	Y_Med << '\n';
	sacc_ << '\n';
	hacc_ << '\n';
	lambdaacc_ << lambdaacc[sample] << '\n';
	etaacc_ << etaacc[sample] << '\n';
      }

      // Delete the arrays we allocated
      delete[] mugave;
      delete[] sigmag2ave;
      delete[] xave;
      delete[] yave;
      delete[] mugacc;
      delete[] taugacc;
      delete[] sacc;
      delete[] hacc;
      delete[] lambdaacc;
      delete[] etaacc;

      // Close the ofstreams (files) we opened
      a_.close();
      b2.close();
      beta_.close();
      mugave_.close();
      sigmag2ave_.close();
      xave_.close();
      yave_.close();
      sacc_.close();
      hacc_.close();
      mugacc_.close();
      taugacc_.close();
      lambdaacc_.close();
      etaacc_.close();

      X_Med.close();
      Y_Med.close();
      Mug_Med.close();
      Sigmag2_Med.close();
    }
  // END DIAGNOSTICS ////////////////////////////
}
