/*
 *  This file is part of BGX, the Bayesian Gene eXpression program.
 *  Copyright (c) 2003-2004  Graeme Ambler <graeme@ambler.me.uk>
 *
 *  BGX is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License, version 2, as
 *  published by the Free Software Foundation.
 *
 *  BGX is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifndef LOCAL_RAND_INCLUDED
#define LOCAL_RAND_INCLUDED

#include <valarray>

#ifdef __INTEL_COMPILER
#include <mathimf.h>
#else
#include <math.h>
#endif

#include <boost/random.hpp>
using boost::uint32_t;

template<class GenType, typename RealType = double> class Rand;
template<class EngineType, typename RealType = double> class Boost_Wrap;
typedef Rand<Boost_Wrap<boost::lagged_fibonacci4423> > Random;
//typedef Rand<Boost_Wrap<boost::mt19937> > Random;

template<class EngineType, typename RealType>
class Boost_Wrap{
private:
  EngineType Engine;
  boost::uniform_real<RealType> U01;
  boost::variate_generator<EngineType, boost::uniform_real<RealType> > Generator;
public:
  Boost_Wrap(uint32_t seed) : Engine(seed), U01(), Generator(Engine,U01) {}
  
  void seed(uint32_t seed) { Engine.seed(seed); }
  
  RealType operator()() { return Generator(); }
};

template<class GenType, typename RealType>
class Rand{
private:
  GenType unif;
public:
  Rand(uint32_t seed=643774788) : unif(seed), GM_a1(0), GM_a2(0) {}

  void seed(uint32_t seed) { unif.seed(seed); }

  RealType Uniform() { return unif(); }

  RealType Uniform(RealType a, RealType b) { return (b-a)*this->Uniform()+a; }

  // Don't use boost::normal_distribution, as it is slower than mine.
  RealType Normal();

  // Note: in Normal(,) (below) we use std.dev. not variance!!!
  RealType Normal(RealType mu, RealType sigma) { return mu+this->Normal()*sigma; }

  // Don't use boost::exponential_distribution, as it is slower than mine.
  RealType Exponential() { return -log(this->Uniform()); }

  RealType Exponential(RealType lambda) { return this->Exponential()/lambda; }

  RealType Gamma(RealType a);

  RealType Gamma(RealType a, RealType b) { return this->Gamma(a)/b; }

  RealType Cauchy() { return tan(M_PI*(this->Uniform()-.5)); }

  RealType Beta(RealType a, RealType b) { RealType x=this->Gamma(a); RealType y=this->Gamma(b); return x/(x+y); }

  // Dirichlet overwrites the input valarray with the output array.
  void Dirichlet(std::valarray<RealType>& alpha, int length);
  
private:
  RealType GM_a1;
  RealType GM_a2;
  RealType GM_s;
  RealType GM_s2;
  RealType GM_d;
  RealType GM_q0;
  RealType GM_b;
  RealType GM_sigma;
  RealType GM_c;
};

#include "rand.cc"

#endif // LOCAL_RAND_INCLUDED
